package edu.iastate.jrelm.rl.rotherev.advanced;

import edu.iastate.jrelm.core.Action;
import edu.iastate.jrelm.similarity.DissimilarityMeasure;
import edu.iastate.jrelm.spillover.SpilloverByDissimilarity;
import edu.iastate.jrelm.spillover.SpilloverWeightingException;


/**
 * Implements a simple logarithmic weight function of the form
 * 
 * 			a * ln(b * (x + c))
 * 
 * where x is a value indicating the dissimilarity of two Actions as determined 
 * by the given DissimilarityMeasure. 
 * <p>
 * Note: The VRELearner uses weight values between 0 and 1. As such, there 
 * are some restrictions in place on the logarithmic function. Parameter 'a' 
 * must be set to a value <= 0 and parameters 'b' and 'c' must be > 0. This is 
 * still not enough to guarantee that all weight values are between 0 and 1, so 
 * the funtion is capped such that values > 1 are reduced to 1 and values less 
 * than zero are set to zero. 
 *  
 * @author Charles Gieseler
 *
 * @param <A> - The type of Action
 * @param <D> - The type of DissimilarityMeasure to use.
 * <p>
 * Note: both D is restricted to DissimilarityMeasures 
 * that return Doubles as indicators.
 */
public class LogarithmicSpillover<
		A extends Action, 
		D extends DissimilarityMeasure<A, Double>> 
		extends SpilloverByDissimilarity<A,D> {

	private double paramA;
	private double paramB;
	private double paramC;
	
	/**
	 * Construct a weighing method with the given 
	 * DissimilarityMeasure and parameter values. 
	 * 
	 * @param measure - the SimilarityMeasure to use for comparison of Actions
	 * @param a - parameter for adjusting the shape of the exponential curve
	 */
	public LogarithmicSpillover(D measure, double a, double b, double c){
		super(measure);
		paramA = a;
		paramB = b;
	}
	
	/**
	 * Generates a reward spillover weight using a logarithmic function of the 
	 * form
	 * 
	 * 			a * ln(b * (x + c))
	 * 
	 * where x is a numerical (Double) indicator of how dissimilar the given Action
	 * is to the reference Action. 

	 */
	public double generateWeight(A action) {
		double weight = 0;
		Double dissimilarityValue;
		double x;
		
		//Compare the given Action with the reference Action to
		//determine how similar they are
		A refAction = getReferenceAction();
		dissimilarityValue = getDissimilarityMeasure().dissimilarity(refAction, action);
		x = dissimilarityValue.doubleValue();
		
		//Calculate the weight value
		//		a * log(b (x + c))
		weight = paramA * Math.log(paramB * (x + paramC));
		
		//Cap the function values to be between 0 and 1
		if(weight < 0)
			weight = 0;
		
		if(weight > 1)
			weight = 1;
		
		return weight;
	}

	public String getName(){return "Logarithmic Weighting";}
	
	/**
	 * Get the value for parameter a. 
	 * 
	 * @return - value of 'a'
	 */
	public double getParamA() {
		return paramA;
	}
	
	/**
	 * Set the value of parameter a.
	 * Note: must be a value <= 0.
	 * @param aValue - value for a
	 */
	public void setParamA(double aValue) throws SpilloverWeightingException {
		if(aValue <= 0)
			this.paramA = aValue;
		else
			throw new SpilloverWeightingException("Illegal value for parameter 'a' in " +
					"LogarithmicSpillover: " + aValue +". Must set 'a' to a value <= 0.");
	}
	
	/**
	 * Get the value for parameter b. 
	 * 
	 * @return - value of 'b'
	 */
	public double getParamB() {
		return paramB;
	}
	
	/**
	 * Set the value of parameter b.
	 * Note: must be a value > 0.
	 * @param paramB
	 */
	public void setParamB(double bValue) throws SpilloverWeightingException {
		if(bValue > 0)
			this.paramB = bValue;
		else
			throw new SpilloverWeightingException("Illegal value for parameter 'b' in " +
					"LogarithmicSpillover: " + bValue +". Must set 'b' to a value > 0.");
	}
	
	/**
	 * Get the value for parameter c. 
	 * 
	 * @return - value of 'c'
	 */
	public double getParamC() {
		return paramC;
	}
	
	/**
	 * Set the value of parameter c.
	 * Note: must be a value > 0.
	 * @param cValue - value for c
	 */
	public void setParamC(double cValue) throws SpilloverWeightingException {
		if(cValue > 0)
			this.paramC = cValue;
		else
			throw new SpilloverWeightingException("Illegal value for parameter 'C' in " +
					"LogarithmicSpillover: " + cValue +". Must set 'c' to a value > 0.");
	}
}
