package edu.iastate.jrelm.spillover;

import edu.iastate.jrelm.core.Action;

/**
 * Some reinforcement learning algorithms may use spillover of rewards in updating
 * their policies. This is where Actions similar to the last Action chosen receive 
 * partial reward values. The proportion of reward that is assigned to a similar 
 * Action is determined by a spillover weight. This class is for implementing 
 * functions to calculate spillover weight. 
 * <p>
 * Spillover occurs for Actions that meet certain simialrity requirements. That is,
 * only Actions that are sufficieantly similar to a fixed reference Action will 
 * recieve partial rewards. As such, child classes will make use of a provided 
 * SimilarityMeasure or DissimilarityMeasure to compare Actions. The reference Action 
 * is usually the last Action chosen and must also be set before a spillover weight 
 * can be calculated.
 * 
 * @author Charles Gieseler
 *
 * @param <A> - the type of Action to compare and generate weights for.
 */
public abstract class SpilloverWeightGenerator<A extends Action> {
	//This is the Action that all other Actions are compared with when 
	//generateWeight() is called. This is usually the last Action chosen
	//by the learning algorithm and can be set through setReferenceAction()
	private A refAction;
	
	/**
	 * Get the fixed Action with which other Actions are compared. 
	 * @return the reference action
	 */
	public A getReferenceAction(){
		return refAction;
	}
	
	/**
	 * Set the fixed Action with which other Actions are compared. 
	 * @param referenc
	 */
	public void setReferenceAction(A referenceAction){
		refAction = referenceAction;
	}
	
	/**
	 * Get the title of this spillover method as a String.
	 * @return - name of the spillover method
	 */
	public abstract String getName();
	
	/**
	 * A spillover weight is the proportion of reward an Action should receive. This
	 * weight is based on how similar the given Action is to a set reference Action. 
	 * Comparison of Actions is either performed by a SimilarityMeasure or a 
	 * DissimilarityMeasure. 
	 * 
	 * @param action - the action to generate a spillover weight for.
	 * @return a new spillover weight
	 */
	public abstract double generateWeight(A action);
}
